import os
import re
from distutils import dir_util

from . import conf


class Error(Exception):
    pass


def _prepare_output_dir(input_d, output_d):
    dir_util.copy_tree(input_d, output_d)


def get_output_id():
    id_ = conf.START_OUTPUT_INDEX
    conf.START_OUTPUT_INDEX += 1
    return id_


def _parse_annotations(file):
    with open(file, "r") as f:
        content = f.read()

    data = {}

    data["inputs"] = __parse_inputs(content)
    data["entrypoints"] = __parse_entrypoints(content)
    content, data["outputs"] = __parse_outputs(content)

    return content, data


def __parse_outputs(input_file):
    outputs = {}
    # check how many outputs have been declared in the input file
    p = re.compile(conf.OUTPUT_PATTERN, re.MULTILINE | re.ASCII)
    outputs_name = list(p.findall(input_file))

    num_outputs = len(outputs_name)

    # replace outputs with the replacement string
    output_file = input_file
    for i in range(num_outputs):
        num = get_output_id()
        output_file = p.sub(conf.OUTPUT_REPL.format(num), output_file, count=1)
        outputs.update({outputs_name[i]: num})

    return output_file, outputs


def __parse_inputs(content):
    return __parse(content, conf.INPUT_PATTERN, conf.START_INPUT_INDEX)


def __parse_entrypoints(content):
    return __parse(content, conf.ENTRY_PATTERN, conf.START_ENTRY_INDEX)


def __parse(content, pattern, start_index):
    p = re.compile(pattern, re.MULTILINE | re.ASCII)
    inputs = p.findall(content)

    return {v: i for (i, v) in enumerate(inputs, start_index)}


def add_io_declarations(file, data):

    with open(file, "r") as f:
        input_file = f.read()

    inputs_str = ""
    entrys_str = ""
    outputs_str = ""

    inputs = list(data["inputs"].keys())
    outputs = list(data["outputs"].keys())
    entrypoints = list(data["entrypoints"].keys())

    for input_ in inputs:
        inputs_str += conf.IO_SIGNATURE.format(input_)

    for output in outputs:
        outputs_str += conf.IO_SIGNATURE.format(output)

    for entrypoint in entrypoints:
        entrys_str += conf.IO_SIGNATURE.format(entrypoint)

    # prepare output file (inject code)
    output_file = input_file
    output_file = output_file.replace("{input_funcs}", inputs_str)
    output_file = output_file.replace("{output_funcs}", outputs_str)
    output_file = output_file.replace("{entry_funcs}", entrys_str)

    # write ta.h
    with open(file, "w") as f:
        f.write(output_file)


def fill_input_array(file, data):

    with open(file, "r") as f:
        input_file = f.read()

    # prepare instructions for filling the `input_funcs` array
    num_inputs = len(data["inputs"])
    array_fill = ",".join(data["inputs"])

    num_entrys = len(data["entrypoints"])
    entrys_fill = ",".join(data["entrypoints"])

    # prepare output file (inject code)
    output_file = input_file
    output_file = output_file.replace("{num_inputs}", str(num_inputs))
    output_file = output_file.replace("{fill_inputs}", array_fill)

    output_file = output_file.replace("{num_entrys}", str(num_entrys))
    output_file = output_file.replace("{fill_entrys}", entrys_fill)

    # write output file
    with open(file, "w") as f:
        f.write(output_file)


def _copy(src, des):
    with open(os.path.join(conf.STUBS_FOLDER, src), "r") as f:
        content = f.read()

    with open(os.path.join(des, src), "w") as f:
        f.write(content)

def _mkdir(path):
    if not os.path.exists(path):
        os.mkdir(path)



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

with open("requirements.txt", "r") as f:
    requirements = f.readlines()

setuptools.setup(
    name="tz-codegen",
    version="0.1.4",
    author="Sepideh Pouyanrad",
    description="Trustzone module code generator for the Authentic Execution framework",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    install_requires=requirements,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: POSIX :: Linux",
    ],
    python_requires='>=3.6',
    entry_points={
        'console_scripts': ['TZ-Code-Gen = tzcodegen.generator:__main']
    },
    include_package_data=True,
)



import logging
import os
import sys
import uuid

from . import conf
from .utils import _parse_annotations, _prepare_output_dir, _copy, _mkdir, \
    add_io_declarations, fill_input_array
from .initialization import _set_parser, _set_logging


def __run(args):

    module_name = os.path.basename(os.path.normpath(args.input))
    out_include = os.path.join(args.output, "include")
    _mkdir(out_include)
    _copy(conf.STUB_PTA_H, out_include)
    _copy(conf.STUB_SPONGENT_H, out_include)
    _copy(conf.STUB_AE_H, out_include)

    _copy(conf.STUB_SPONGENT, args.output)
    _copy(conf.STUB_Makefile, args.output)
    # -------------------------------user_ta_header_defines-------------------------------
    _copy(conf.STUB_USER_TA_HEADER_DEFINES, args.output)
    u = uuid.uuid4()
    int_uuid = u.int
    
    n = [', 0x'] * 11
    n[::2] = ['{:012x}'.format(u.node)[i:i + 2] for i in range(0, 12, 2)]

    inject_uuid = "\t{ " + "0x{:08x}".format(u.time_low) + ", " \
        + "0x{:04x}".format(u.time_mid) + ", " + "0x{:04x}".format(u.time_hi_version) \
        + ", {" + "0x{:02x}".format(u.clock_seq_hi_variant)\
        + ", " + "0x{:02x}".format(u.clock_seq_low) + \
        ", " + "0x" + "".join(n) + "} }"

    with open(os.path.join(args.output, conf.STUB_USER_TA_HEADER_DEFINES), "r") as f:
        content = f.read()

    content = content.replace("{module_uuid}", inject_uuid)

    with open(os.path.join(args.output, conf.STUB_USER_TA_HEADER_DEFINES), "w") as f:
        f.write(content)
    # -----------------------------sub.mk----------------------------------------------
    _copy(conf.STUB_SUB_MK, args.output)
    with open(os.path.join(args.output, conf.STUB_SUB_MK), "r") as f:
        content = f.read()

    content = content.replace("{module_name}", module_name+".c")

    with open(os.path.join(args.output, conf.STUB_SUB_MK), "w") as f:
        f.write(content)

    # ---------------------------module----------------------------------------------
    out_src = os.path.join(args.output, module_name+".c")

    # In this section, we update TA:
    # - parse the annotations (inputs, outputs, entry points)
    # - add imports
    # parse annotations
    content, data = _parse_annotations(out_src)

    # read imports
    with open(os.path.join(conf.STUBS_FOLDER, conf.STUB_IMPORT), "r") as f:
        import_str = f.read()

    # write new content
    new_content = import_str + "\n" + "#include " + \
        "<" + module_name + ".h>\n" + content
    with open(out_src, "w") as f:
        f.write(new_content)
    # ------------------creating ta.h file--------------------------------------
    with open(os.path.join(conf.STUBS_FOLDER, conf.STUB_TA_H), "r") as f:
        ta_h = f.read()

    with open(os.path.join(out_include, module_name + ".h"), "w") as f:
        f.write(ta_h)

    add_io_declarations(os.path.join(out_include, module_name + ".h"), data)
    # ------------------------------------------------------------------
    ## Authentic Execution file ##
    # In this section, we add to the project all the needed for authentic execution
    # we also need to update some data structures with the information retrieved
    # before

    with open(os.path.join(conf.STUBS_FOLDER, conf.STUB_AUTH_EXEC), "r") as f:
        auth_exec = f.read()

    auth_exec = auth_exec.replace("{header_file}", "<" + module_name + ".h>")

    with open(os.path.join(args.output, conf.STUB_AUTH_EXEC), "w") as f:
        f.write(auth_exec)

    fill_input_array(os.path.join(args.output, conf.STUB_AUTH_EXEC), data)

    return data, int_uuid


def generate(args):
    try:
        # check if the input dir is a correct Rust Cargo module
        logging.debug("Checking input project..")
        #cargo = _check_input_module(args.input)

        # copy dir
        logging.debug("Creating output project..")
        _prepare_output_dir(args.input, args.output)
        # generate code
        logging.debug("Generating code..")
        return __run(args)

    except Exception as e:
        logging.error(e)
        sys.exit(1)


def __main():
    # argument parser
    parser = _set_parser()
    args = parser.parse_args()

    # set logging params
    _set_logging(args.loglevel)
    generate(args)


if __name__ == "__main__":
    __main()



import os

DEFAULT_LOG_LEVEL = "info"
STUBS_FOLDER = os.path.join(os.path.dirname(
    os.path.abspath(__file__)), "stubs")


# Stubs
STUB_PTA_H = "pta_attestation.h"
STUB_SPONGENT_H = "spongent.h"
STUB_AE_H = "authentic_execution.h"
STUB_SPONGENT = "spongent.c"
STUB_Makefile = "Makefile"
STUB_USER_TA_HEADER_DEFINES = "user_ta_header_defines.h"
STUB_SUB_MK = "sub.mk"
STUB_IMPORT = "import.c"
STUB_TA_H = "ta.h"
STUB_AUTH_EXEC = "authentic_execution.c"

# Starting entrypoint index
# 0 is set_key, 1 is attest, 2 is handle_input
START_ENTRY_INDEX = 3
# Starting indexes of inputs, outputs
# They need to have different indexes, because the `index` field in Connection does
# not distinguish between them. If the same index is used for different types, bad
# things can happen. Moreover, having these "ranges" allow us do identify what is
# an index: e.g., 25848 is an output, 44 is an input, etc.
# We believe 16384 values for each type is more than enough for a single module.
START_INPUT_INDEX = 0
START_OUTPUT_INDEX = 16384

OUTPUT_PATTERN = '^[ \t]*SM_OUTPUT[ \t]*\([ \t]*([a-zA-Z_][a-zA-Z_0-9]*)[ \t]*\)[ \t]*;[ \t]*$'
INPUT_PATTERN = ('^[ \t]*SM_INPUT[ \t]*\([ \t]*([a-zA-Z_][a-zA-Z_0-9]*)[ \t]*,'
                 '[ \t]*[a-zA-Z_][a-zA-Z_0-9]*[ \t]*,[ \t]*[a-zA-Z_][a-zA-Z_0-9]*[ \t]*\)')
ENTRY_PATTERN = ('^[ \t]*SM_ENTRY[ \t]*\([ \t]*([a-zA-Z_][a-zA-Z_0-9]*)[ \t]*,'
                 '[ \t]*[a-zA-Z_][a-zA-Z_0-9]*[ \t]*,[ \t]*[a-zA-Z_][a-zA-Z_0-9]*[ \t]*\)')
OUTPUT_REPL = "SM_OUTPUT_AUX(\\1, {});"
IO_SIGNATURE = ("void {}(void *session, uint32_t param_types, TEE_Param params[4],"
                "unsigned char *data, uint32_t len);\n")
